"""
libyt-specific data structures



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import importlib
import os
import time
import weakref
from collections import defaultdict

import numpy as np
import yt.frontends.api
from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import mylog, setdefaultattr
from yt.geometry.geometry_handler import YTDataChunk
from yt.geometry.grid_geometry_handler import GridIndex

from .fields import libytFieldInfo


class libytGrid(AMRGridPatch):
    # We will set _id_offset to libyt.param_yt["index_offset"]
    # when initializing libytHierarchy
    _id_offset = 0

    def __init__(self, id, index, level):
        AMRGridPatch.__init__(self, id, filename=None, index=index)

        # Might be redundant by setting these two, as they refer to the same thing.
        # MPI_rank : Refers to grid is currently belongs to which MPI rank.
        # filename : Refers to grid is currently belongs to which MPI rank. See line 101.
        self.MPI_rank = None
        self.Parent = None
        self.Children = []
        self.Level = level

    def __repr__(self):
        return "libytGrid_%09i (dimension = %s)" % (self.id, self.ActiveDimensions)


class libytHierarchy(GridIndex):
    grid = libytGrid
    libyt = None
    # _preload_implemented = True # Not sure about this option

    def __init__(self, ds, dataset_type="libyt"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        self.directory = os.getcwd()
        self.float_type = "float64"
        self.libyt = self.dataset.libyt

        libytGrid._id_offset = self.libyt.param_yt["index_offset"]

        GridIndex.__init__(self, ds, dataset_type)

    def _detect_output_fields(self):
        if "field_list" in self.libyt.param_yt:
            field_list = self.libyt.param_yt["field_list"]
            self.field_list = [
                (self.libyt.param_yt["frontend"].lower(), v) for v in field_list.keys()
            ]
        else:
            mylog.debug("No field list \"libyt.param_yt['field_list']\".")

        if "particle_list" in self.libyt.param_yt:
            particle_list = self.libyt.param_yt["particle_list"]
            for ptype in particle_list.keys():
                attribute = particle_list[ptype]["attribute"]
                self.field_list += [(ptype, particle) for particle in attribute.keys()]
        else:
            mylog.debug("No particle list \"libyt.param_yt['particle_list']\".")

    def _count_grids(self):
        self.num_grids = self.libyt.param_yt["num_grids"]

    def _initialize_grid_arrays(self):
        # We don't need to initialize a new buffer for hierarchy,
        # since these array are initialized and set in libyt.
        pass

    def _parse_index(self):
        mylog.debug("Setting index array for %s grids.", self.num_grids)

        # get hierarchy information of all grids from libyt
        hierarchy = self.libyt.hierarchy
        self.grid_dimensions = hierarchy["grid_dimensions"]
        self.grid_left_edge = self.ds.arr(hierarchy["grid_left_edge"], "code_length")
        self.grid_right_edge = self.ds.arr(hierarchy["grid_right_edge"], "code_length")
        self.grid_levels = hierarchy["grid_levels"]

        # derive max_level
        self.max_level = self.grid_levels.max()

        # derive grid_particle_count from par_count_list.
        # par_count_list is created only if there is particles.
        if "par_count_list" in hierarchy:
            self.grid_particle_count = np.sum(hierarchy["par_count_list"], axis=1)
            self.grid_particle_count = self.grid_particle_count[..., np.newaxis]
        else:
            self.grid_particle_count = np.zeros((self.num_grids, 1), "int32")

        # Indicates which MPI rank it belongs to.
        self.proc_num = hierarchy["proc_num"]

        # allocate all grid objects
        self.grids = np.empty(self.num_grids, dtype="object")
        index_offset = self.libyt.param_yt["index_offset"]
        for index in range(self.num_grids):
            self.grids[index] = self.grid(index + index_offset, self, self.grid_levels.flat[index])
            self.grids[index].MPI_rank = self.proc_num[index, 0]
            self.grids[index].filename = "MPI_Rank_%07i" % (self.proc_num[index, 0])

    def _populate_grid_objects(self):
        # must flat it since 'grid_parent_id' has the dimension [num_grids][1]
        parent_list = self.libyt.hierarchy["grid_parent_id"].flat
        index_offset = self.libyt.param_yt["index_offset"]

        for index in range(self.num_grids):
            grid = self.grids[index]
            parent_id = parent_list[index]

            # set up the parent-children relationship if parent grid exist
            if parent_id >= index_offset:
                parent_grid = self.grids[parent_id - index_offset]
                grid.Parent = parent_grid
                parent_grid.Children.append(grid)

            # set up other grid attributes
            grid._prepare_grid()
            grid._setup_dx()

    def _chunk_io(self, dobj, cache=True, local_only=False):
        gfiles = defaultdict(list)
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for g in gobjs:
            gfiles[g.filename].append(g)

        for fn in sorted(gfiles):
            if local_only:
                gobjs = [g for g in gfiles[fn] if g.MPI_rank == self.comm.rank]
                gfiles[fn] = gobjs
            gs = gfiles[fn]
            count = self._count_selection(dobj, gs)
            yield YTDataChunk(dobj, "io", gs, count, cache=cache)


class libytDataset(Dataset):
    _index_class = libytHierarchy
    _field_info_class = libytFieldInfo
    # _field_info_base_class = object
    _dataset_type = (
        "libyt"  # must set here since __init__() does not know dataset_type when calling it
    )
    _debug = False  # debug mode for libyt (not supported yet), some checks are in libyt C-library. Cannot open yet. Future use.
    libyt = None

    def __init__(self, units_override=None, unit_system="cgs"):
        # nothing to do if initialization has been done
        if self.libyt is not None:
            return

        # load the libyt module
        self.libyt = self._obtain_libyt()

        # get the target frontend (precisely speaking, the target Dataset subclass)
        # and set fluid type and fields accordingly
        self._code_frontend = self.libyt.param_yt["frontend"].lower()
        found_frontend = False
        for name in yt.frontends.api._frontends:
            if self._code_frontend == name.lower():
                # Import frontend dataset
                # The naming rule of dataset class is XXXDataset, XXX doesn't necessarily need to be all capital
                frontend = importlib.import_module(f"yt.frontends.{name.lower()}.api")
                frontend_dataset = None
                for _ in dir(frontend):
                    if _.endswith("Dataset"):
                        frontend_dataset = getattr(frontend, _)

                # Borrow frontend's field info
                self._field_info_class = frontend_dataset._field_info_class
                self.fluid_types += (self._code_frontend,)
                found_frontend = True

                # Read from param_yt['field_list'] and update the XXXDataset._field_info_class.
                # This action accumulates in each round, because we change the class
                # static variable in XXXDataset._field_info_class.
                if "field_list" in self.libyt.param_yt:
                    field_list = self.libyt.param_yt["field_list"]
                    known_other_fields = list(self._field_info_class.known_other_fields)
                    for field_name in field_list.keys():
                        # Step1 : Check if field_name is already inside known_other_fields
                        field_exist = False
                        for index in range(len(known_other_fields)):
                            if field_name == known_other_fields[index][0]:
                                field_exist = True
                                # Step2 : If field_name exists, append alias names one by one if it's not in the name
                                # list yet
                                for name_alias in field_list[field_name]["attribute"][1]:
                                    if name_alias not in known_other_fields[index][1][1]:
                                        known_other_fields[index][1][1].append(name_alias)
                                break
                        # Step2 : If field_name doesn't exist in known_other_fields, add a new field to add_fields list
                        if field_exist is False:
                            known_other_fields.append(
                                (field_name, tuple(field_list[field_name]["attribute"]))
                            )

                    # Step3 : convert it back to tuple
                    self._field_info_class.known_other_fields = tuple(known_other_fields)
                else:
                    mylog.debug("No self.libyt.param_yt['field_list'].")

                # Read from param_yt['particle_list'] and update XXXDataset._field_info_class.
                if "particle_list" in self.libyt.param_yt:
                    particle_list = self.libyt.param_yt["particle_list"]
                    known_particle_fields = list(self._field_info_class.known_particle_fields)
                    for ptype in particle_list.keys():
                        attribute = particle_list[ptype]["attribute"]
                        for particle in attribute.keys():
                            par_exist = False
                            for index in range(len(known_particle_fields)):
                                if particle == known_particle_fields[index][0]:
                                    par_exist = True
                                    for name_alias in attribute[particle][1]:
                                        if name_alias not in known_particle_fields[index][1][1]:
                                            known_particle_fields[index][1][1].append(name_alias)
                                    break

                            if par_exist is False:
                                known_particle_fields.append((particle, tuple(attribute[particle])))

                    self._field_info_class.known_particle_fields = tuple(known_particle_fields)
                else:
                    mylog.debug("No self.libyt.param_yt['particle_list'].")

                break
        if found_frontend is False:
            # We assume that user's code has corresponding yt frontend, if not, terminate yt.
            raise NotImplementedError(
                "libyt set frontend = %s, cannot find the code frontend [ %sDataset ] in yt."
                % (self.libyt.param_yt["frontend"], self.libyt.param_yt["frontend"].upper())
            )

        mylog.info("libyt: code dataset       = libytDataset")
        mylog.info(f"libyt: FieldInfo subclass = {self._field_info_class}")
        mylog.info(f"libyt: fluid type         = {self._code_frontend}")

        Dataset.__init__(
            self,
            self.libyt.param_yt["fig_basename"],
            self._dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _set_code_unit_attributes(self):
        # currently libyt assumes cgs
        setdefaultattr(self, "length_unit", self.quan(self.libyt.param_yt["length_unit"], "cm"))
        setdefaultattr(self, "mass_unit", self.quan(self.libyt.param_yt["mass_unit"], "g"))
        setdefaultattr(self, "time_unit", self.quan(self.libyt.param_yt["time_unit"], "s"))
        setdefaultattr(
            self, "magnetic_unit", self.quan(self.libyt.param_yt["magnetic_unit"], "gauss")
        )

    def _parse_parameter_file(self):
        # dataset identifier
        self.unique_identifier = time.time()

        # user-specific parameters
        self.parameters.update(self.libyt.param_user)

        # yt-specific parameters
        param_yt = self.libyt.param_yt
        self.current_time = param_yt["current_time"]
        self.dimensionality = param_yt["dimensionality"]
        self.refine_by = param_yt["refine_by"]
        self.cosmological_simulation = param_yt["cosmological_simulation"]
        self.current_redshift = param_yt["current_redshift"]
        self.omega_matter = param_yt["omega_matter"]
        self.omega_lambda = param_yt["omega_lambda"]
        self.hubble_constant = param_yt["hubble_constant"]

        # vectors are stored as tuples in libyt and must be converted to NumPy arrays
        self.domain_left_edge = np.asarray(param_yt["domain_left_edge"])
        self.domain_right_edge = np.asarray(param_yt["domain_right_edge"])
        self.domain_dimensions = np.asarray(param_yt["domain_dimensions"])
        self._periodicity = (
            bool(param_yt["periodicity"][0] == 1),
            bool(param_yt["periodicity"][1] == 1),
            bool(param_yt["periodicity"][2] == 1),
        )

        # Load code specific parameters
        for key in self.libyt.param_user.keys():
            if hasattr(self, key):
                mylog.info(
                    "Overwrite existing attribute self.%s = %s in class libytDataset",
                    key,
                    getattr(self, key),
                )

            setattr(self, key, self.libyt.param_user[key])
            mylog.info(
                "Set attribute self.%s = %s in class libytDataset.",
                key,
                self.libyt.param_user[key],
            )

    @staticmethod
    def _obtain_libyt():
        import libyt

        from ._version import __version__

        libyt_version = f"{libyt.libyt_info['version'][0]}.{libyt.libyt_info['version'][1]}.{libyt.libyt_info['version'][2]}"
        mylog.info(f"libyt    version = {libyt_version}")
        mylog.info(f"yt_libyt version = {__version__}")

        return libyt

    @classmethod
    def _is_valid(cls, *args, **kwargs):
        # always return false since libyt is used for inline analysis only
        return False



"""
libyt-specific IO functions



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

import numpy as np
from yt.funcs import mylog
from yt.geometry.selection_routines import AlwaysSelector
from yt.utilities.io_handler import BaseIOHandler


class libytIOHandler(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "libyt"

    def __init__(self, ds):
        super(libytIOHandler, self).__init__(ds)
        import libyt

        self.libyt = libyt
        self.ds = ds
        self.grid_data = libyt.grid_data
        self.param_yt = libyt.param_yt
        self.hierarchy = libyt.hierarchy
        self._field_dtype = "float64"
        self.myrank = libytIOHandler._get_my_rank()

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)

        # Get position (coordinate) label.
        ptf_new = {}
        for ptype in ptf.keys():
            coor_label = self.param_yt["particle_list"][ptype]["particle_coor_label"]
            if None in coor_label:
                raise ValueError("Particle label representing postion X/Y/Z not set!")
            ptf_new[ptype] = coor_label

        # Get remote data.
        nonlocal_data = self._prepare_remote_particle_from_libyt(chunks, ptf_new)

        # Get index offset and particle data dict
        index_offset = self.param_yt["index_offset"]
        particle_data = self.libyt.particle_data

        for chunk in chunks:
            for g in chunk.objs:
                # fetch the position x/y/z of particle by ptype
                for ptype in ptf.keys():
                    # Get particle count in ptype, continue if it is zero
                    index_label = self.param_yt["particle_list"][ptype]["label"]
                    if self.hierarchy["par_count_list"][g.id - index_offset][index_label] == 0:
                        continue

                    coor_label = self.param_yt["particle_list"][ptype]["particle_coor_label"]
                    if g.MPI_rank == self.myrank:
                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[0] in particle_data[g.id][ptype]
                        ):
                            x = particle_data[g.id][ptype][coor_label[0]]
                        else:
                            x = self.libyt.get_particle(g.id, ptype, coor_label[0])

                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[1] in particle_data[g.id][ptype]
                        ):
                            y = particle_data[g.id][ptype][coor_label[1]]
                        else:
                            y = self.libyt.get_particle(g.id, ptype, coor_label[1])

                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[2] in particle_data[g.id][ptype]
                        ):
                            z = particle_data[g.id][ptype][coor_label[2]]
                        else:
                            z = self.libyt.get_particle(g.id, ptype, coor_label[2])
                    else:
                        x = nonlocal_data[g.id][ptype][coor_label[0]]
                        y = nonlocal_data[g.id][ptype][coor_label[1]]
                        z = nonlocal_data[g.id][ptype][coor_label[2]]

                    # g.id ptype particle number is 0, libyt.get_particle will return None,
                    # It will not happen unless something went wrong when passing particle count to libyt.
                    if x is None or y is None or z is None:
                        raise ValueError("Particle position should not be None.")
                    else:
                        yield ptype, (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)

        # Get position (coordinate) label and append particle attribute to get after them.
        ptf_new = {}
        for ptype in ptf.keys():
            coor_label = self.param_yt["particle_list"][ptype]["particle_coor_label"].copy()
            if None in coor_label:
                raise ValueError("Particle label representing postion X/Y/Z not set!")
            ptf_new[ptype] = coor_label
            for field in ptf[ptype]:
                ptf_new[ptype].append(field)
            ptf_new[ptype] = set(ptf_new[ptype])

        # Get remote data.
        nonlocal_data = self._prepare_remote_particle_from_libyt(chunks, ptf_new)

        # Get index offset and particle data
        index_offset = self.param_yt["index_offset"]
        particle_data = self.libyt.particle_data

        for chunk in chunks:
            for g in chunk.objs:
                # fetch particle data.
                for ptype in ptf.keys():
                    # get particle count in ptype, continue if it is zero
                    index_label = self.param_yt["particle_list"][ptype]["label"]
                    if self.hierarchy["par_count_list"][g.id - index_offset][index_label] == 0:
                        continue

                    # fetch the position x/y/z of particle by ptype
                    coor_label = self.param_yt["particle_list"][ptype]["particle_coor_label"]
                    if None in coor_label:
                        raise ValueError("Particle label representing postion X/Y/Z not set!")
                    if g.MPI_rank == self.myrank:
                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[0] in particle_data[g.id][ptype]
                        ):
                            x = particle_data[g.id][ptype][coor_label[0]]
                        else:
                            x = self.libyt.get_particle(g.id, ptype, coor_label[0])

                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[1] in particle_data[g.id][ptype]
                        ):
                            y = particle_data[g.id][ptype][coor_label[1]]
                        else:
                            y = self.libyt.get_particle(g.id, ptype, coor_label[1])

                        if (
                            g.id in particle_data
                            and ptype in particle_data[g.id]
                            and coor_label[2] in particle_data[g.id][ptype]
                        ):
                            z = particle_data[g.id][ptype][coor_label[2]]
                        else:
                            z = self.libyt.get_particle(g.id, ptype, coor_label[2])
                    else:
                        x = nonlocal_data[g.id][ptype][coor_label[0]]
                        y = nonlocal_data[g.id][ptype][coor_label[1]]
                        z = nonlocal_data[g.id][ptype][coor_label[2]]

                    # g.id ptype particle number is 0, libyt.get_particle will return None.
                    # It will not happen unless something went wrong when passing particle count to libyt.
                    if x is None or y is None or z is None:
                        raise ValueError("Particle position should not be None.")

                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        continue

                    for field in ptf[ptype]:
                        if g.MPI_rank == self.myrank:
                            if (
                                g.id in particle_data
                                and ptype in particle_data[g.id]
                                and field in particle_data[g.id][ptype]
                            ):
                                data = particle_data[g.id][ptype][field]
                            else:
                                data = self.libyt.get_particle(g.id, ptype, field)
                        else:
                            data = nonlocal_data[g.id][ptype][field]

                        # if ptype particle num in grid g.id = 0, get_particle will return None.
                        # It will not happen unless something went wrong when passing particle count to libyt.
                        if data is None:
                            raise ValueError("Particle data should not be None.")
                        else:
                            yield (ptype, field), data[mask]

    def _read_chunk_data(self, chunk, fields):
        # TODO: The suite hasn't been tested yet.
        #       Although it's be use for caching, I wonder do libyt need this.
        #       Since we don't need to load data from file. Although we do need
        #       to get data from remote rank.
        rv = {}
        if len(chunk.objs) == 0:
            return rv
        for g in chunk.objs:
            rv[g.id] = {}

        # Split into particles and non-particles
        fluid_fields, particle_fields = [], []
        for ftype, fname in fields:
            if ftype in self.ds.particle_types:
                particle_fields.append((ftype, fname))
            else:
                fluid_fields.append((ftype, fname))

        # Read particle data
        if len(particle_fields) > 0:
            selector = AlwaysSelector(self.ds)
            rv.update(self._read_particle_selection([chunk], selector, particle_fields))

        # If no more fluid fields to read, return rv. Or else, read fluid fields
        if len(fluid_fields) == 0:
            return rv

        # Prepare nonlocal data
        nonlocal_data = self._prepare_remote_field_from_libyt([chunk], fields)

        # TODO: Bug, rv should allocate a new buffer.
        for field in fluid_fields:
            ftype, fname = field
            for g in chunk.objs:
                if g.MPI_rank == self.myrank:
                    rv[g.id][field] = self._get_field_from_libyt(g, fname)
                else:
                    rv[g.id][field] = self._get_field_from_libyt(
                        g, fname, nonlocal_data=nonlocal_data
                    )
        return rv

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        chunks = list(chunks)

        # Prepare nonlocal data
        nonlocal_data = self._prepare_remote_field_from_libyt(chunks, fields)

        # TODO: Allocate buffer for rv, don't make rv point directly to simulation data buffer.
        # if selector.__class__.__name__ == "GridSelector":
        #     if not (len(chunks) == len(chunks[0].objs) == 1):
        #         raise RuntimeError("class libytIOHandler, def _read_fluid_selection, selector == GridSelector, "
        #                            "chunk to be read not equal to 1.")
        #     g = chunks[0].objs[0]
        #     for ftype, fname in fields:
        #         if g.MPI_rank == self.myrank:
        #             rv[(ftype, fname)] = self._get_field_from_libyt(g, fname)
        #         else:
        #             rv[(ftype, fname)] = self._get_field_from_libyt(g, fname, nonlocal_data=nonlocal_data)
        #     return rv

        if size is None:
            size = sum((g.count(selector) for chunk in chunks for g in chunk.objs))

        for field in fields:
            rv[field] = np.empty(size, dtype=self._field_dtype)

        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids", size, [f2 for f1, f2 in fields], ng
        )

        # Get grid data
        for field in fields:
            offset = 0
            ftype, fname = field
            for chunk in chunks:
                for g in chunk.objs:
                    if g.MPI_rank == self.myrank:
                        data_view = self._get_field_from_libyt(g, fname)
                    else:
                        data_view = self._get_field_from_libyt(
                            g, fname, nonlocal_data=nonlocal_data
                        )
                    offset += g.select(selector, data_view, rv[field], offset)
            assert offset == size
        return rv

    @staticmethod
    def _get_my_rank():
        from mpi4py import MPI

        comm = MPI.COMM_WORLD
        return comm.Get_rank()

    def _distinguish_nonlocal_grids(self, chunks):
        # Split local and non-local grids.
        from mpi4py import MPI

        comm = MPI.COMM_WORLD

        local_id = []
        nonlocal_id = []
        nonlocal_rank = []

        for chunk in chunks:
            for g in chunk.objs:
                if g.MPI_rank != self.myrank:
                    nonlocal_id.append(g.id)
                    nonlocal_rank.append(g.MPI_rank)
                else:
                    local_id.append(g.id)
        num_nonlocal_grids = len(nonlocal_id)

        # Gather all non-local grids in each rank.
        sendcounts = comm.gather(num_nonlocal_grids, root=0)
        sendcounts = comm.bcast(sendcounts, root=0)
        sendbuf = np.asarray(nonlocal_id)
        recvbuf = np.empty(sum(sendcounts), dtype=int)
        comm.Gatherv(sendbuf=sendbuf, recvbuf=(recvbuf, sendcounts), root=0)
        comm.Bcast(recvbuf, root=0)

        # Get grid id that this rank has to prepare, grid id doesn't have to be 0-indexed
        index_offset = self.param_yt["index_offset"]
        proc_num = self.hierarchy["proc_num"][:, 0]
        index = np.argwhere(proc_num[recvbuf - index_offset] == self.myrank)
        to_prepare = list(np.unique(recvbuf[index]))

        # Determine whether we should call for libyt C extend method for RMA operation.
        if sum(sendcounts) != 0:
            rma = True
        else:
            rma = False

        return rma, to_prepare, nonlocal_id, nonlocal_rank

    def _prepare_remote_field_from_libyt(self, chunks, fields):
        # Wrapper for the RMA operation at libyt C library code.
        # Each rank must call this method, in order to get nonlocal grids.

        # Distinguish local and non-local grid, and what should this rank prepared.
        rma, to_prepare, nonlocal_id, nonlocal_rank = self._distinguish_nonlocal_grids(chunks)

        if rma is True:
            # Encode field name to UTF-8
            fname_list = []
            for _ftype, fname in fields:
                fname_list.append(fname.encode(encoding="UTF-8", errors="strict"))
            fname_list = sorted(set(fname_list))

            # Get nonlocal_data, libyt will perform RMA operation in this step.
            # Every rank must call this libyt method.
            mylog.debug("Getting nonlocal data through libyt ...")
            nonlocal_data = self.libyt.get_field_remote(
                fname_list,
                len(fname_list),
                to_prepare,
                len(to_prepare),
                nonlocal_id,
                nonlocal_rank,
                len(nonlocal_id),
            )
        else:
            nonlocal_data = None

        return nonlocal_data

    def _prepare_remote_particle_from_libyt(self, chunks, ptf):
        # Wrapper for the RMA operation at libyt C library code.
        # For supporting particles. Each rank must call this method.

        # Distinguish local and non-local grid, and what should this rank prepared.
        rma, to_prepare, nonlocal_id, nonlocal_rank = self._distinguish_nonlocal_grids(chunks)

        if rma is True:
            # Filter out those who really has particles in their grid. Grid id doesn't have to be 0-indexed.
            # Since we aren't sure how many particle types will yt access, we check total particle counts.
            index_offset = self.param_yt["index_offset"]
            par_count = self.ds.index.grid_particle_count[:, 0]

            # If to_prepare is empty list, we don't need to filter out grids without particles
            if len(to_prepare) != 0:
                to_prepare = np.asarray(to_prepare)
                index = np.argwhere(par_count[to_prepare - index_offset] > 0)
                to_prepare = list(to_prepare[index].flatten())

            # If nonlocal_id / nonlocal_rank is empty list, we don't need to filter out grids without particles
            if len(nonlocal_id) != 0:
                nonlocal_id = np.asarray(nonlocal_id)
                index = np.argwhere(par_count[nonlocal_id - index_offset] > 0)
                nonlocal_id = list(nonlocal_id[index].flatten())
                nonlocal_rank = np.asarray(nonlocal_rank)
                nonlocal_rank = list(nonlocal_rank[index].flatten())

            # String inside ptf should be encoded in UTF-8, and attributes should be in list obj.
            ptf_c = {}
            for key in ptf.keys():
                attr_list = []
                for attr in ptf[key]:
                    attr_list.append(attr.encode(encoding="UTF-8", errors="strict"))
                ptype = key.encode(encoding="UTF-8", errors="strict")
                ptf_c[ptype] = sorted(set(attr_list))

            # Call libyt RMA
            mylog.debug("Getting nonlocal data through libyt ...")
            nonlocal_data = self.libyt.get_particle_remote(
                ptf_c,
                ptf_c.keys(),
                to_prepare,
                len(to_prepare),
                nonlocal_id,
                nonlocal_rank,
                len(nonlocal_id),
            )
        else:
            nonlocal_data = None

        return nonlocal_data

    def _get_field_from_libyt(self, grid, fname, nonlocal_data=None):
        # This method is to get the grid data.
        # If nonlocal_data is none, which means to get a local grid.
        # Otherwise, read the nonlocal data in nonlocal_data.
        field_list = self.param_yt["field_list"]
        ghost_cell = field_list[fname]["ghost_cell"]
        if field_list[fname]["field_type"] == "cell-centered":
            # Read data from grid_data, or nonlocal_data.
            # We don't create key-value pair if no data pass in from user.
            try:
                if nonlocal_data is None:
                    data_convert = self.grid_data[grid.id][fname]
                else:
                    data_convert = nonlocal_data[grid.id][fname]
            except Exception as err:
                mylog.error("%s: %s", type(err).__name__, str(err))
                mylog.error(
                    "Cannot get cell-centered grid [%s] data on MPI rank [%d]."
                    % (grid.id, grid.MPI_rank)
                )
                raise RuntimeError("libyt didn't get the data successfully.")

            # Remove ghost cell, and get my slice
            data_shape = data_convert.shape
            data_convert = data_convert[
                ghost_cell[0] : (data_shape[0] - ghost_cell[1]),
                ghost_cell[2] : (data_shape[1] - ghost_cell[3]),
                ghost_cell[4] : (data_shape[2] - ghost_cell[5]),
            ]

        elif field_list[fname]["field_type"] == "face-centered":
            # Read data from grid_data, or nonlocal_data.
            # We don't create key-value pair if no data pass in from user.
            try:
                if nonlocal_data is None:
                    data_temp = self.grid_data[grid.id][fname]
                else:
                    data_temp = nonlocal_data[grid.id][fname]
            except Exception as err:
                mylog.error("%s: %s", type(err).__name__, str(err))
                mylog.error(
                    "Cannot get face-centered grid [%s] data on MPI rank [%d]."
                    % (grid.id, grid.MPI_rank)
                )
                raise RuntimeError("libyt didn't get the data successfully.")

            # Remove ghost cell, and get my slice
            data_shape = data_temp.shape
            data_temp = data_temp[
                ghost_cell[0] : (data_shape[0] - ghost_cell[1]),
                ghost_cell[2] : (data_shape[1] - ghost_cell[3]),
                ghost_cell[4] : (data_shape[2] - ghost_cell[5]),
            ]

            # Convert to cell-centered
            grid_dim = self.hierarchy["grid_dimensions"][grid.id]
            if field_list[fname]["contiguous_in_x"] is True:
                grid_dim = np.flip(grid_dim)
            axis = np.argwhere(grid_dim != data_temp.shape).flatten()
            if len(axis) != 1 or data_temp.shape[axis[0]] - 1 != grid_dim[axis[0]]:
                mylog.error(
                    "Field [%s] in grid [%d] is not a face-centered data. "
                    "It has dim = %s, but it should be dim = %s"
                    % (fname, grid.id, (data_temp.shape,), grid_dim)
                )
                raise ValueError("Face-centered data dimension not match.")

            if axis == 0:
                data_convert = 0.5 * (data_temp[:-1, :, :] + data_temp[1:, :, :])
            elif axis == 1:
                data_convert = 0.5 * (data_temp[:, :-1, :] + data_temp[:, 1:, :])
            elif axis == 2:
                data_convert = 0.5 * (data_temp[:, :, :-1] + data_temp[:, :, 1:])

        elif field_list[fname]["field_type"] == "derived_func":
            # Read data
            try:
                if nonlocal_data is None:
                    data_convert = self.libyt.derived_func(grid.id, fname)
                else:
                    data_convert = nonlocal_data[grid.id][fname]
            except Exception as err:
                mylog.error("%s: %s", type(err).__name__, str(err))
                mylog.error(
                    "Cannot get derived field data in grid [%s] on MPI rank [%d]."
                    % (grid.id, grid.MPI_rank)
                )
                raise RuntimeError("libyt didn't get the data successfully.")
        else:
            # Since we only supports "cell-centered", "face-centered", "derived_func" tags for now
            # Raise an error if enter this block.
            raise ValueError(
                f"libyt does not have field_type [ {field_list[fname]['field_type']} ]"
            )

        # Swap axes or not, then return
        if field_list[fname]["contiguous_in_x"] is True:
            return data_convert.swapaxes(0, 2)
        else:
            return data_convert



"""
API for yt.frontends.libyt



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Top-level package for yt_libyt."""

from ._version import __version__
from .data_structures import libytDataset, libytGrid, libytHierarchy
from .fields import libytFieldInfo
from .io import libytIOHandler



"""
libyt-specific fields



"""

# -----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# -----------------------------------------------------------------------------

from yt.fields.field_info_container import FieldInfoContainer


# this is the FieldInfo subclass adopted when libyt fails to find a matched code frontend
# For now, we assume that users already have their own frontends. So this class shouldn't be called.
class libytFieldInfo(FieldInfoContainer):
    known_other_fields = ()
    known_particle_fields = ()

    def __init__(self, ds, field_list):
        super(libytFieldInfo, self).__init__(ds, field_list)

    def setup_fluid_fields(self):
        # currently, this is not in use.
        pass

    def setup_particle_fields(self, ptype):
        super(libytFieldInfo, self).setup_particle_fields(ptype)



import pytest


def test_import():
    import yt_libyt

    api_list = ["libytDataset", "libytGrid", "libytHierarchy", "libytFieldInfo", "libytIOHandler"]
    for attr in api_list:
        assert hasattr(yt_libyt, attr) is True, f"No class named {attr} in yt_libyt"


def test_always_pass():
    assert True


@pytest.mark.xfail
def test_expected_failure():
    assert 1 == 2
